/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DebuggerMain.h"
#include "DebugEventServer.h"
#include "Configuration.h"
#include "ByteStreamRegion.h"
#include "ByteShiftArray.h"
#include "SymbolManager.h"
#include "DebuggerPluginInterface.h"
#include "DialogAttach.h"
#include "DialogMemoryRegions.h"
#include "Util.h"
#include "DisassemblyLib.h"
#include "i386ArchProcessor.h"
#include "Expression.h"
#include "CommentServer.h"
#include "version.h"
#include "RecentFileManager.h"

#include <QSettings>
#include <QToolButton>
#include <QMessageBox>
#include <QFileDialog>
#include <QInputDialog>
#include <QTableWidgetItem>
#include <QFileInfo>
#include <QTextStream>

// TODO: move elf parsing functionality to another class
#include <elf.h>

//------------------------------------------------------------------------------
// Name: DebuggerMain(QWidget *parent)
// Desc: constructor
//------------------------------------------------------------------------------
DebuggerMain::DebuggerMain(QWidget *parent) : QMainWindow(parent), 
		m_DebugEventServer(0), m_UpdateGUI(true), m_ReenableBreakPointAddress(0), 
		m_LastEvent(0), m_StepWasRun(false), m_LastOpenDirectory(""), 
		m_ArchProcessor(0), m_CommentServer(0), m_RecentFileManager(0), 
		m_StackViewLocked(false) {

	ui.setupUi(this);
	
	// setup the global pointers as early as possible
	Debugger::debuggerGUI = this;
	Debugger::setDebugEventHandler(this);

	// make the stack view all nice..
	// TODO: remember what the user set it as, and restore it!
	ui.stackView->setRowWidth(1);
	ui.stackView->setWordWidth(4);
	ui.stackView->setShowAsciiDump(false);
	
	// setup tabs
	setupTabButtons();
	
	// set the listbox to about 4 lines
	QFontMetrics fm(ui.listWidget->font());
	ui.listWidget->setFixedHeight(fm.height() * 5);
	
	// register our DebugEvent type so it can be used in 
	// inter-thread signals
	qRegisterMetaType<DebugEvent>("DebugEvent");
	
	// make sure our widgets use custom context menus
	ui.registerList->setContextMenuPolicy(Qt::CustomContextMenu);
	ui.cpuView->setContextMenuPolicy(Qt::CustomContextMenu);
	ui.stackView->setContextMenuPolicy(Qt::CustomContextMenu);
			
	// add toggles for the dock windows
	ui.menu_View->addAction(ui.registersDock->toggleViewAction());
	ui.menu_View->addAction(ui.dataDock->toggleViewAction());
	ui.menu_View->addAction(ui.stackDock->toggleViewAction());
	ui.menu_View->addAction(ui.toolBar->toggleViewAction());

	// remove the one in the designer and put in our built one
	// it's a bit ugly to do it this way, but the designer wont let me 
	// make a tabless entry..and the ones i create look slightly diff 
	// (probably lack of layout manager in mine...
	ui.tabWidget->removeTab(0);
	createNewDataTab();
	
	doApplyDefaultFonts();
	
	m_StackViewInfo.view = ui.stackView;
	
	// enable the arch processor for i386
	m_ArchProcessor = new i386ArchProcessor;
	m_ArchProcessor->setupRegisterView(ui.registerList);
	m_ArchProcessor->updateRegisterView("");
	
	// setup the comment server for the stack viewer
	m_CommentServer = new CommentServer;
	ui.stackView->setCommentServer(m_CommentServer);
	
	// default the working directory to ours
	m_WorkingDirectory = QDir().absolutePath();
	
	m_RecentFileManager = new RecentFileManager;
	ui.action_Recent_Files->setMenu(m_RecentFileManager->createMenu());
	connect(m_RecentFileManager, SIGNAL(fileSelected(const QString &)), this, SLOT(doOpenFile(const QString &)));
}


//------------------------------------------------------------------------------
// Name: ~DebuggerMain()
// Desc: destructor
//------------------------------------------------------------------------------
DebuggerMain::~DebuggerMain() {

	switch(Debugger::config()->closeBehaviour) {
	case Configuration::Terminate:
		on_action_Kill_triggered();
		break;
	case Configuration::Detach:
		on_action_Detach_triggered();
		break;
	}
	
	delete m_CommentServer;
	delete m_ArchProcessor;
	delete m_RecentFileManager;
	
	// remove the last tab
	removeDataTab();
	
	// clean up the stream for the stack view too
	safeDelete(m_StackViewInfo.stream);
}

//------------------------------------------------------------------------------
// Name: doApplyDefaultFonts()
// Desc: applies the confirations fonts to all neccessary widgets
//------------------------------------------------------------------------------
void DebuggerMain::doApplyDefaultFonts() {

	QFont font;

	// set some default fonts
	if(font.fromString(Debugger::config()->StackViewFont)) {
		ui.stackView->setFont(font);
	}
	
	if(font.fromString(Debugger::config()->RegisterViewFont)) {
		ui.registerList->setFont(font);
	}

	if(font.fromString(Debugger::config()->DisassemblyViewFont)) {
		ui.cpuView->setFont(font);
	}
	
	if(font.fromString(Debugger::config()->MemoryDumpViewFont)) {
		foreach(DataViewInfo *dataView, m_DataRegions) {
			dataView->view->setFont(font);
		}
	}
}

//------------------------------------------------------------------------------
// Name: finishPluginSetup(QVector<DebuggerPluginInterface *> &plugins)
// Desc: finalizes plugin setup by adding each to the menu, we can do this now 
//       that we have a GUI widget to attach it to
//------------------------------------------------------------------------------
void DebuggerMain::finishPluginSetup(QVector<DebuggerPluginInterface *> &plugins) {
	// setup the menu for all plugins that which to do so
	foreach (DebuggerPluginInterface *plugin, plugins) {
		QMenu *const menu = plugin->menu(this);
		if(menu != 0) {
			ui.menu_Plugins->addMenu(menu);
		}
	}
}

//------------------------------------------------------------------------------
// Name: closeEvent(QCloseEvent *)
// Desc: triggered on main window close, saves window state
//------------------------------------------------------------------------------
void DebuggerMain::closeEvent(QCloseEvent *) {
	saveWindowState();
}

//------------------------------------------------------------------------------
// Name: showEvent(QShowEvent *)
// Desc: triggered on show, restores window state
//------------------------------------------------------------------------------
void DebuggerMain::showEvent(QShowEvent *) {
	loadWindowState();
}

//------------------------------------------------------------------------------
// Name: loadWindowState()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::loadWindowState() {
	QSettings settings;
	settings.beginGroup("Window");
	const QByteArray state = settings.value("window_state").toByteArray();
	
	const int width = settings.value("window_width", -1).toInt();
	const int height = settings.value("window_height", -1).toInt();
	if(width != -1) {
		resize(width, size().height());
	}
	
	if(height != -1) {
		resize(size().width(), height);
	}
	
	settings.endGroup();
	restoreState(state);
}

//------------------------------------------------------------------------------
// Name: saveWindowState()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::saveWindowState() {
	QSettings settings;
	const QByteArray state = saveState();
	settings.beginGroup("Window");
	settings.setValue("window_state", state);
	settings.setValue("window_width", width());
	settings.setValue("window_height", height());
	settings.endGroup();
}

//------------------------------------------------------------------------------
// Name: setupTabButtons()
// Desc: creates the add/remove tab buttons in the data view
//------------------------------------------------------------------------------
void DebuggerMain::setupTabButtons() {
	// add the corner widgets to the data view
	m_AddTab = new QToolButton(ui.tabWidget);
	m_DelTab = new QToolButton(ui.tabWidget);
	
	m_AddTab->setToolButtonStyle(Qt::ToolButtonIconOnly);
	m_DelTab->setToolButtonStyle(Qt::ToolButtonIconOnly);
	
	m_AddTab->setIcon(QIcon(":/debugger/pics/addtab.png"));
	m_DelTab->setIcon(QIcon(":/debugger/pics/closetab.png"));
	m_AddTab->setAutoRaise(true);
	m_DelTab->setAutoRaise(true);
	m_DelTab->setEnabled(false);
	
	ui.tabWidget->setCornerWidget(m_AddTab, Qt::TopLeftCorner);
	ui.tabWidget->setCornerWidget(m_DelTab, Qt::TopRightCorner);
	
	connect(m_AddTab, SIGNAL(clicked()), this, SLOT(createNewDataTab()));
	connect(m_DelTab, SIGNAL(clicked()), this, SLOT(removeDataTab()));
}

//------------------------------------------------------------------------------
// Name: on_registerList_customContextMenuRequested(const QPoint &pos)
// Desc: context menu handler for register view
//------------------------------------------------------------------------------
void DebuggerMain::on_registerList_customContextMenuRequested(const QPoint &pos) {
	QTreeWidgetItem *const item =  ui.registerList->itemAt(pos);
	if(item != 0) {
		if(!ui.registerList->isCatagory(item)) {
			QMenu *menu = new QMenu(this);
			menu->addAction("&Follow In Dump", this, SLOT(followInDump()));
			menu->addAction("&Follow In Dump (New Tab)", this, SLOT(followInDumpNewTab()));
			menu->addAction("&Follow In Stack", this, SLOT(followInStack()));
			menu->exec(ui.registerList->mapToGlobal(pos));
			safeDelete(menu);
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_cpuView_breakPointToggled(BaseTypes::address_t address)
// Desc: handler for toggling the breakpoints
//------------------------------------------------------------------------------
void DebuggerMain::on_cpuView_breakPointToggled(BaseTypes::address_t address) {
	Debugger::toggleBreakPoint(address);
}

//------------------------------------------------------------------------------
// Name: on_registerList_itemDoubleClicked(QTreeWidgetItem *item)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_registerList_itemDoubleClicked(QTreeWidgetItem *item) {

	const QString s = item->text(0);
	BaseTypes::reg_t *value = valueFromName(s, State::REG_ANY);
	
	if(value != 0) {
		if(Debugger::getValueFromUser(*value, "Register Value")) {			
			Debugger::debuggerBase->setState(Debugger::currentState);
			updateGUI();
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_action_About_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_About_triggered() {

	QMessageBox::about(this, "About EDB", QString().sprintf(
		"<p>EDB (Evan's Debugger) is designed to be an easy to use, modular, and cross platform debugger.</p>"
		"<p>More information and updates can be found at <a href=\"http://www.codef00.com\">http://www.codef00.com</a></p>"
		"<p>Written by Evan Teran.</p>"
		"<p>version: %s</p>", version_string));
}

//------------------------------------------------------------------------------
// Name: on_actionAbout_QT_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionAbout_QT_triggered() {
	QMessageBox::aboutQt(this, "About QT");
}

//------------------------------------------------------------------------------
// Name: on_action_Configure_Debugger_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Configure_Debugger_triggered() {
	
	if(Debugger::dialogOptions()->exec() == QDialog::Accepted) {	

		// reload symbols incase they changed, or our symbol files changes
		safeDelete(Debugger::symbolManager);
		Debugger::symbolManager = new SymbolManager(Debugger::config()->symbolPath);

		if(Debugger::memoryRegions != 0) {
			Debugger::memoryRegions->sync();
		}
		
		doApplyDefaultFonts();
		updateGUI();
	}
}

//------------------------------------------------------------------------------
// Name: on_cpuView_customContextMenuRequested(const QPoint &pos)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_cpuView_customContextMenuRequested(const QPoint &pos) {
	QMenu *menu = new QMenu(this);
	// TODO: add comment
	//menu->addAction("Add &Comment", this, SLOT(""));
	//menu->addSeparator();
	menu->addAction("&Goto Address", this, SLOT(mnuCPUJumpToAddress()));
	menu->addAction("&Goto EIP", this, SLOT(mnuCPUJumpToEIP()));
	
	menu->exec(ui.cpuView->viewport()->mapToGlobal(pos));
	safeDelete(menu);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
BaseTypes::address_t DebuggerMain::getGotoExpression(bool &ok) {
    QString text = QInputDialog::getText(this, "Goto Address", "Address:", QLineEdit::Normal, "", &ok);
	
	if (ok && !text.isEmpty()) {
		Expression<BaseTypes::address_t> expr(text, &Debugger::getVariable, 0, &Debugger::getDWORD, 0);
		try {
			const BaseTypes::address_t address = expr.evaluateExpression();
			ok = true;
			return address;
		} catch(const Error &err) {
			QMessageBox::information(this, "Error In Address Expression!", err.what());
			ok = false;
		}
	}
	return 0;
}



//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doMemoryFollow(bool ok, BaseTypes::address_t address, bool (*followFunc)(BaseTypes::address_t)) {
	if(ok) {
		if(!(*followFunc)(address)) {
			QMessageBox::information(this, "No Memory Found", QString().sprintf("There appears to be no memory at that location (<strong>%p</strong>)", reinterpret_cast<void *>(address)));
		}
	}
}

//------------------------------------------------------------------------------
// Name: followRegisterInDump(bool tabbed)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::followRegisterInDump(bool tabbed) {
	bool ok;
	const BaseTypes::address_t address = getFollowRegister(ok);
	if(ok) {
		if(!Debugger::dumpData(address, tabbed)) {
			QMessageBox::information(this, "No Memory Found", QString().sprintf("There appears to be no memory at that location (<strong>%p</strong>)", reinterpret_cast<void *>(address)));
		}
	}
}

//------------------------------------------------------------------------------
// Name: mnuGotoESPStack()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuGotoESPStack() {
	doMemoryFollow(true, Debugger::currentState.esp, Debugger::dumpStack);
}

//------------------------------------------------------------------------------
// Name: mnuGotoEBPStack()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuGotoEBPStack() {
	doMemoryFollow(true, Debugger::currentState.ebp, Debugger::dumpStack);
}

//------------------------------------------------------------------------------
// Name: mnuCPUJumpToEIP()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUJumpToEIP() {
	doMemoryFollow(true, Debugger::currentState.eip, Debugger::jumpToAddress);
}

//------------------------------------------------------------------------------
// Name: mnuCPUJumpToAddress()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuCPUJumpToAddress() {
    bool ok;
	const BaseTypes::address_t address = getGotoExpression(ok);
	doMemoryFollow(true, address, Debugger::jumpToAddress);
}

//------------------------------------------------------------------------------
// Name: mnuGotoAddress()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuGotoAddress() {
    bool ok;
	const BaseTypes::address_t address = getGotoExpression(ok);
	doMemoryFollow(ok, address, Debugger::dumpData);
}

//------------------------------------------------------------------------------
// Name: mnuGotoAddressStack()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuGotoAddressStack() {
    bool ok;
	const BaseTypes::address_t address = getGotoExpression(ok);
	doMemoryFollow(ok, address, Debugger::dumpStack);
}

//------------------------------------------------------------------------------
// Name: followInStack()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::followInStack() {
	bool ok;
	const BaseTypes::address_t address = getFollowRegister(ok);
	doMemoryFollow(ok, address, Debugger::dumpStack);
}

//------------------------------------------------------------------------------
// Name: doFollowInStack(const QHexView *hv)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doFollowInStack(const QHexView *hv) {
	bool ok;
	const BaseTypes::address_t address = getFollowAddress(hv, ok);
	doMemoryFollow(ok, address, Debugger::dumpStack);
}

//------------------------------------------------------------------------------
// Name: doFollowInDump(const QHexView *hv)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doFollowInDump(const QHexView *hv) {
	bool ok;
	const BaseTypes::address_t address = getFollowAddress(hv, ok);
	doMemoryFollow(ok, address, Debugger::dumpData);
}

//------------------------------------------------------------------------------
// Name: getFollowRegister(bool &ok)
// Desc: 
//------------------------------------------------------------------------------
BaseTypes::reg_t DebuggerMain::getFollowRegister(bool &ok) {
	ok = false;
	const QTreeWidgetItem *const i = ui.registerList->currentItem();
	BaseTypes::reg_t ret = 0;
	if(i != 0) {
		const QString s = i->text(0);
		BaseTypes::reg_t *const p = valueFromName(s, State::REG_GPR | State::REG_IP);
		if(p != 0) {
			ok = true;
			ret = *p;
		}
	}
	
	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
BaseTypes::address_t DebuggerMain::getFollowAddress(const QHexView *hv, bool &ok) {
	ok = false;
	
	if(hv != 0) {
		if(hv->hasSelectedText()) {
			const QByteArray data = hv->selectedBytes();

			if(data.size() == sizeof(BaseTypes::address_t)) {
				union {
					quint8					b[sizeof(BaseTypes::address_t)];
					BaseTypes::address_t	d;
				} dword;

				for(unsigned int i = 0; i < sizeof(BaseTypes::address_t); ++i) {
					dword.b[i] = data[i];
				}

				ok = true;
				return dword.d;
			} else {
				QMessageBox::information(this, "Nothing Selected", QString().sprintf("Please select a %d bytes to use this function.", sizeof(BaseTypes::address_t)));
			}
		} else {
			QMessageBox::information(this, "Nothing Selected", QString().sprintf("Please select a %d bytes to use this function.", sizeof(BaseTypes::address_t)));
		}
	}
	return 0;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuFollowInDump() {
	doFollowInDump(qobject_cast<QHexView *>(ui.tabWidget->currentWidget()));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuFollowInStack() {
	doFollowInStack(qobject_cast<QHexView *>(ui.tabWidget->currentWidget()));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackFollowInDump() {
	doFollowInDump(ui.stackView);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::mnuStackFollowInStack() {
	doFollowInStack(ui.stackView);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionApplication_Arguments_triggered() {
    bool ok;
    const QString text = QInputDialog::getText(this, "Application Arguments", "Arguments:", QLineEdit::Normal, m_Arguments, &ok);
	if(ok) {
		m_Arguments = text;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionApplication_Working_Directory_triggered() {
    bool ok;
    const QString text = QInputDialog::getText(this, "Application Working Directory", "Working Directory:", QLineEdit::Normal, m_WorkingDirectory, &ok);
	if(ok) {
		m_WorkingDirectory = text;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::pushDWORD() {

	BaseTypes::reg_t value;
	
	Debugger::debuggerBase->readBytes(Debugger::currentState.esp - sizeof(value), &value, sizeof(value));

	if(Debugger::getValueFromUser(value, "Enter value to push")) {

		Debugger::currentState.esp -= sizeof(value);
		Debugger::debuggerBase->writeBytes(Debugger::currentState.esp, &value, sizeof(value));

		Debugger::debuggerBase->setState(Debugger::currentState);
		updateGUI();
	}
	
}


//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::popDWORD() {
	Debugger::currentState.esp += sizeof(BaseTypes::reg_t);
	updateGUI();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_stackView_customContextMenuRequested(const QPoint &pos) {
	QHexView *const s = ui.stackView;
	if(s != 0) {
		QMenu *menu = s->createStandardContextMenu();
		menu->addSeparator();
		menu->addAction("Follow Address In &Dump", this, SLOT(mnuStackFollowInDump()));
		menu->addAction("Follow Address In &Stack", this, SLOT(mnuStackFollowInStack()));
		menu->addAction("&Goto Address", this, SLOT(mnuGotoAddressStack()));
		menu->addAction("Goto E&SP", this, SLOT(mnuGotoESPStack()));
		menu->addAction("Goto E&BP", this, SLOT(mnuGotoEBPStack()));
		menu->addSeparator();
		menu->addAction("&Edit Bytes", this, SLOT(modifyBytesStack()));
		menu->addSeparator();
		menu->addAction("&Push DWORD", this, SLOT(pushDWORD()));
		menu->addAction("P&op DWORD", this, SLOT(popDWORD()));
		menu->exec(s->mapToGlobal(pos));
		safeDelete(menu);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::dumpContextMenu(const QPoint &pos) {
	QHexView *const s = qobject_cast<QHexView *>(sender());
	if(s != 0) {
		QMenu *menu = s->createStandardContextMenu();
		menu->addSeparator();
		menu->addAction("Follow Address In &Dump", this, SLOT(mnuFollowInDump()));
		menu->addAction("Follow Address In &Stack", this, SLOT(mnuFollowInStack()));
		menu->addAction("&Goto Address", this, SLOT(mnuGotoAddress()));
		menu->addSeparator();
		menu->addAction("&Edit Bytes", this, SLOT(modifyBytesDump()));
		menu->addSeparator();
		menu->addAction("&Dump To File", this, SLOT(dumpDataToFile()));
		menu->exec(s->mapToGlobal(pos));
		safeDelete(menu);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::dumpDataToFile() {
	QHexView *const s = qobject_cast<QHexView *>(ui.tabWidget->currentWidget());
	if(s != 0) {
		
		const QString filename = QFileDialog::getSaveFileName(this, "Save File", m_LastOpenDirectory, "");
		if(!filename.isEmpty()) {
			
			QFile file(filename);
			file.open(QIODevice::WriteOnly);
			if(file.isOpen()) {
				file.write(s->allBytes());
			}			
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doModifyBytes(QHexView *s) {
	if(s != 0) {
		const BaseTypes::address_t address = s->selectedBytesAddress();
		const unsigned int size = s->selectedBytesSize();
		QByteArray bytes = s->selectedBytes();
		
		
		if(Debugger::getBinaryStringFromUser(bytes, "Edit Binary String", size)) {
		
			// fill with 0x00 bytes
			// TODO: make this value choosable
			while(bytes.size() < static_cast<int>(size)) {
				bytes.push_back('\x00');
			}
		
			Debugger::debuggerBase->writeBytes(address, bytes.data(), size);
			updateGUI();
		}	
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::modifyBytesDump() {
	doModifyBytes(qobject_cast<QHexView *>(ui.tabWidget->currentWidget()));
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::modifyBytesStack() {
	doModifyBytes(ui.stackView);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateMenuState(GUI_STATE guiState) {

	switch(guiState) {
	case PAUSED:
	#if 0
		ui.actionRun_Until_Return->setEnabled(true);
	#endif
		ui.action_Restart->setEnabled(true);
		ui.action_Run->setEnabled(true);
		ui.action_Pause->setEnabled(false);
		ui.action_Step_Into->setEnabled(true);
		ui.action_Step_Over->setEnabled(true);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(true);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(true);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(true);
		ui.action_Detach->setEnabled(true);
		ui.action_Kill->setEnabled(true);
		break;
	case RUNNING:
	#if 0
		ui.actionRun_Until_Return->setEnabled(false);
	#endif
		ui.action_Restart->setEnabled(false);
		ui.action_Run->setEnabled(false);
		ui.action_Pause->setEnabled(true);
		ui.action_Step_Into->setEnabled(false);
		ui.action_Step_Over->setEnabled(false);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Detach->setEnabled(true);
		ui.action_Kill->setEnabled(true);
		break;
	case TERMINATED:
	#if 0
		ui.actionRun_Until_Return->setEnabled(false);
	#endif
		ui.action_Restart->setEnabled(false);
		ui.action_Run->setEnabled(false);
		ui.action_Pause->setEnabled(false);
		ui.action_Step_Into->setEnabled(false);
		ui.action_Step_Over->setEnabled(false);
		ui.action_Step_Into_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Step_Over_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Run_Pass_Signal_To_Application->setEnabled(false);
		ui.action_Detach->setEnabled(false);
		ui.action_Kill->setEnabled(false);
		break;
	}
	
	m_GUIState = guiState;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventSignaled() {
	switch(m_LastEvent.exitSignal()) {
	case DebugEvent::sigkill:
		QMessageBox::information(this, "Application Killed", "The debugged application was killed.");
		break;
	default:
		Debugger::log(QString().sprintf("Unknown Termination Signal Recieved! %d", m_LastEvent.exitSignal()));
		break;
	}

	on_action_Detach_triggered();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventStopped() {


	// ok we just came in from a signal, we need to test some things, 
	// generally, we will want to check if it was a step, if it was, was it because
	// we just hit a break point or because we wanted to run, but had to step first
	// in case were were on a breakpoint already...

	switch(m_LastEvent.stopSignal()) {
	case DebugEvent::sigsegv:
		QMessageBox::information(this, "Illegal Access Fault", QString().sprintf(
			"<p>The debugged application encountered a segmentation fault!<br />"
			"The address <strong>%p</strong> could not be accessed.</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>", 
			reinterpret_cast<void *>(Debugger::debuggerBase->getFaultAddress()))
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigill:
		QMessageBox::information(this, "Illegal Instruction Fault", 
			"<p>The debugged application attempted to execute an illegal instruction.</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>"
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigfpe:
		QMessageBox::information(this, "Floating Point Exception", 
			"<p>The debugged application encountered a floating-point exception!</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>"
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigabrt:
		QMessageBox::information(this, "Abort", 
			"<p>The debugged application has aborted!</p>"
			"<p>If you would like to pass this exception to the application press Shift+[F7/F8/F9]</p>"
			);
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigstop:
		// user asked to pause the debugged process
		m_StepWasRun = false;
		break;
		
	case DebugEvent::sigtrap:
		
		// we just got a SIGTRAP, there are a few possible causes
		// #1: we just did an execve, in which case we want to stop cause we just attached
		// #2: we hit a 0xCC breakpoint, if so, then we want to stop
		// #3: we did a step
		
		do {
			// prevEIP will have the address of the instruction which brought us here!
			const BaseTypes::address_t prevEIP = Debugger::currentState.eip - Debugger::debuggerBase->executionBreakPointSize();
			
			// look it up in our breakpoint list, make sure it is one of OUR int3s!
			// if it is, we need to backup EIP and pause ourselves
			BreakPoint *const bp = Debugger::debuggerBase->findBreakPoint(prevEIP);
			if(bp != 0 && bp->enabled()) {
						
				// if it's a one time breakpoint then we should remove it upon
				// triggering, this is mainly used for "run-until" situations
				// like step over

				if(bp->oneTime()) {
					Debugger::debuggerBase->removeBreakpoint(prevEIP);
				} else {
					++bp->passcount;					
				}

				// back up eip the size of a breakpoint, since we executed a breakpoint 
				// instead of the real code that belongs there
				Debugger::currentState.eip = prevEIP;
				Debugger::debuggerBase->setState(Debugger::currentState);
				

				// handle conditional breakpoints
				if(m_StepWasRun && !bp->condition.isEmpty()) {
				
					int conditionVal = 0;
					Expression<BaseTypes::address_t> expr(bp->condition, &Debugger::getVariable, 0, &Debugger::getDWORD, 0);
					try {
						conditionVal = expr.evaluateExpression();
					} catch(const Error &err) {
						QMessageBox::information(this, "Error In Breakpoint Expression!", err.what());
						conditionVal = 1;
						// TODO: remove the breakpoint
					}				
					
					// ok, the condition was false, so skip over this breakpoint...
					if(!conditionVal) {
						doResume(0);
						return;
					}
				}
				m_StepWasRun = false;
			} else {
				// it wasn't one of our int3s, so what do we do?

				// this code attempts to cancel a run which will
				// cause overrun a breakpoint potentially				
				uint8_t breakpointData[Debugger::debuggerBase->executionBreakPointSize()];
				if(Debugger::debuggerBase->readBytes(prevEIP, breakpointData, sizeof(breakpointData))) {
					if(std::memcmp(
						breakpointData, 
						Debugger::debuggerBase->executionBreakPointData().data(), 
						Debugger::debuggerBase->executionBreakPointSize()) == 0) {
						m_StepWasRun = false;			
					}
				}
			}
		} while(0);
		break;

	case DebugEvent::sigchld:
		Debugger::debuggerBase->resume(m_LastEvent.stopSignal());
		return;
		
	case DebugEvent::sigcont:
		// this one is probably not very interesting, so we'll just resume
		Debugger::debuggerBase->resume(m_LastEvent.stopSignal());
		// we do an early return so it's like it never happened
		return;
		
	default:
		Debugger::log(QString().sprintf("Unknown Stop Signal Recieved! %d", m_LastEvent.stopSignal()));
		Debugger::debuggerBase->resume(m_LastEvent.stopSignal());
		return;
	}

	// re-enable any breakpoints we disabled
	Debugger::debuggerBase->enableBreakpoint(m_ReenableBreakPointAddress);
	m_ReenableBreakPointAddress = 0;

	// if we stepped with the intention of running, well we should run then!
	if(m_StepWasRun) {
		//m_StepWasRun = false;
		Debugger::debuggerBase->resume(0);
	} else {
		updateGUI();
		updateMenuState(PAUSED);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEventExited() {
	QMessageBox::information(this, "Application Exited", QString().sprintf("The debugged application exited normally with exit code %d.",  m_LastEvent.exitStatus()));
	on_action_Detach_triggered();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::handleEvent(DebugEvent event) {
	
	if(Debugger::debuggerBase != 0) {
		Debugger::debuggerBase->getState(Debugger::currentState);
	}
	
	// it is possible but rare that that we get an event when no memory regions
	// are setup (on detach)
	if(Debugger::memoryRegions != 0) {
		Debugger::memoryRegions->sync();
	}
	
	m_LastEvent = event;
	
	switch(m_LastEvent.reason()) {
	case DebugEvent::EVENT_STOPPED:
		handleEventStopped();
		break;
		
	case DebugEvent::EVENT_SIGNALED:
		handleEventSignaled();
		break;
		
	case DebugEvent::EVENT_EXITED:
		handleEventExited();
		break;
		
	case DebugEvent::EVENT_UNKNOWN:
		break;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::debugEventHandler(DebugEvent event) {
	Q_ASSERT(Debugger::debugEventHandler != 0);
	Debugger::debugEventHandler->handleEvent(event);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::DataViewInfo::update() {

	safeDelete(stream);
	stream = new ByteStreamRegion(region);
	
	if(view != 0) {
		view->setAddressOffset(region.start);
		view->setData(stream);
	}
}


//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateData(DebuggerMain::DataViewInfo *v) {
	if(v != 0) {
		v->update();	
		ui.tabWidget->setTabText(ui.tabWidget->indexOf(v->view), QString().sprintf("%08x-%08x", v->region.start, v->region.end));
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::clearData(DataViewInfo *v) {
	if(v != 0) {
		v->view->clear();
		v->view->scrollTo(0);
		ui.tabWidget->setTabText(ui.tabWidget->indexOf(v->view), QString().sprintf("%08x-%08x", v->region.start, v->region.end));
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doJumpToAddress(BaseTypes::address_t address, const MemoryRegions::Region &r, bool scrollTo) {

	ui.cpuView->setAddressOffset(r.start);
	ui.cpuView->setRegion(r);
	if(scrollTo || !ui.cpuView->addressShown(address)) {
		ui.cpuView->scrollTo(address);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateDisassembly(BaseTypes::address_t address, const MemoryRegions::Region &r) {

	ui.cpuView->setCurrentAddress(address);
	
	doJumpToAddress(address, r, false);
	
	ui.listWidget->clear();
	ui.listWidget->addItems(m_ArchProcessor->updateInstructionInfo(address));
	
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateStackView() {
	if(Debugger::memoryRegions == 0 || !Debugger::dumpStack(Debugger::currentState.esp, !m_StackViewLocked)) {
		ui.stackView->clear();
		ui.stackView->scrollTo(0);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateCPUView(MemoryRegions::Region &region) {
	if(Debugger::memoryRegions == 0 || !Debugger::memoryRegions->findRegion(Debugger::currentState.eip, region)) {
		ui.cpuView->clear();
		ui.cpuView->scrollTo(0);
		ui.listWidget->clear();
	} else {
		updateDisassembly(Debugger::currentState.eip, region);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::updateDataViews() {
	if(Debugger::memoryRegions != 0) {
		// update all data views with the current region data
		foreach(DataViewInfo *info, m_DataRegions) {

			// make sure the regions are still valid..
			if(Debugger::memoryRegions->findRegion(info->region.start)) {
				updateData(info);
			} else {
				clearData(info);
			}
		}
	} else {
		// clear all data views
		foreach(DataViewInfo *info, m_DataRegions) {
			clearData(info);
		}
	}
}

//------------------------------------------------------------------------------
// Name: updateGUI()
// Desc: updates all the different displays
//------------------------------------------------------------------------------
void DebuggerMain::updateGUI() {

	if(m_UpdateGUI) {
	
		MemoryRegions::Region region;

		updateCPUView(region);
		updateDataViews();
		updateStackView();
		m_ArchProcessor->updateRegisterView(region.name);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doResume(bool passExceptionToApp) {
	
	// if we are on a breakpoint, disable it, and set the reenable 
	// address to the current eip
	// disableBreakpoint returns 0x00000000 if there was no breakpoint to disable
	m_ReenableBreakPointAddress = Debugger::debuggerBase->disableBreakpoint(Debugger::currentState.eip);

	// set the state to "running"
	updateMenuState(RUNNING);
	
	// if neccessary pass the signal to the application, otherwise just step
	if(passExceptionToApp && m_LastEvent.stopped() && m_LastEvent.stopSignal() != SIGTRAP && m_LastEvent.stopSignal() != SIGSTOP) {
		Debugger::debuggerBase->step(m_LastEvent.stopSignal());
	} else {
		Debugger::debuggerBase->step(0);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Run_Pass_Signal_To_Application_triggered() {
	m_StepWasRun = true;
	doResume(true);	
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Into_Pass_Signal_To_Application_triggered() {
	doResume(true);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Over_Pass_Signal_To_Application_triggered() {
	doStepOver<
		&DebuggerMain::on_action_Run_Pass_Signal_To_Application_triggered, 
		&DebuggerMain::on_action_Step_Into_Pass_Signal_To_Application_triggered>();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Over_triggered() {
	doStepOver<
		&DebuggerMain::on_action_Run_triggered, 
		&DebuggerMain::on_action_Step_Into_triggered>();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Run_triggered() {
	m_StepWasRun = true;
	doResume(false);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Step_Into_triggered() {
	doResume(false);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_actionRun_Until_Return_triggered() {
	// TODO: implement this
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Pause_triggered() {
	if(Debugger::debuggerBase != 0) {
		Debugger::debuggerBase->pause();
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Restart_triggered() {

	if(Debugger::debuggerBase != 0) {
	
		m_WorkingDirectory = Debugger::getProcessCurrentWorkingDirectory();	
		QStringList args = Debugger::getProcessArguments();
		
		if(args.size() != 0) {
		
			const QString s = args[0];
			args.removeFirst();

			if(!s.isEmpty()) {
				doDetachFromProcess(true);

				if(Debugger::debuggerBase->open(s, m_WorkingDirectory, args)) {
					setInitialDebuggerState();
					setInitialBreakpoint(s);
					setWindowTitle(QString().sprintf("EDB - %s", qPrintable(s)));
				} else {
					QMessageBox::information(this, "Could Not Open", "Failed to open and attach to process, please check priviledges and try again.");
				}

				Debugger::debuggerBase->getState(Debugger::currentState);
				updateGUI();
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::cleanupEventThread() {

	if(m_DebugEventServer != 0) {
		
		// make sure we don't get any spurious events!
		disconnect(m_DebugEventServer, SIGNAL(debugEvent(DebugEvent)), this, SLOT(debugEventHandler(DebugEvent)));
		
		// make sure it thinks it should exit
		m_DebugEventServer->m_ShouldExit = true;

		// this is ugly, but it seems to work pretty well
		// basically we have marked the thread as "should exit"
		// but, we may be blocked (likley) waiting for the next event
		// so we intentionally cause one last event giving it a change to
		// check the loop invariant one more time, which will allow
		// the thread to exit cleanly
		if(m_GUIState == RUNNING) {
			Debugger::debuggerBase->pause();
			updateMenuState(PAUSED);
		} else {
			Debugger::debuggerBase->step(0);
			updateMenuState(RUNNING);
		}
		m_DebugEventServer->wait();
		m_DebugEventServer = 0;
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::cleanupDebugger() {
	
	cleanupEventThread();

	safeDelete(Debugger::memoryRegions);
	safeDelete(Debugger::symbolManager);

	// clear up the data view
	while(ui.tabWidget->count() > 1) {
		removeDataTab();
	}
	
	setWindowTitle("EDB");
	
	updateGUI();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doDetachFromProcess(bool kill) {

	cleanupDebugger();
	
	m_StepWasRun = false;
	
	if(Debugger::debuggerBase != 0) {
		if(kill) {
			Debugger::debuggerBase->kill();
		} else {
			Debugger::debuggerBase->detach();
		}
	}
	
	Debugger::currentState.clear();
	
	updateMenuState(TERMINATED);
}

//------------------------------------------------------------------------------
// Name: on_action_Detach_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Detach_triggered() {
	doDetachFromProcess(false);
}

//------------------------------------------------------------------------------
// Name: on_action_Kill_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Kill_triggered() {
	doDetachFromProcess(true);
}

//------------------------------------------------------------------------------
// Name: getELFEntryPoint(const QString &s)
// Desc: 
//------------------------------------------------------------------------------
BaseTypes::address_t DebuggerMain::getELFEntryPoint(const QString &s) {

	union ElfHeader {
		Elf32_Ehdr elf32;
		Elf64_Ehdr elf64;
	};
	
	QFile file(s);
	file.open(QIODevice::ReadOnly);
	if(file.isOpen()) {
		ElfHeader header;
		file.read(reinterpret_cast<char *>(&header), sizeof(header));
		if(std::memcmp(header.elf32.e_ident, ELFMAG, SELFMAG) == 0) {
			switch(header.elf32.e_ident[EI_CLASS]) {
			case ELFCLASS32:
				return header.elf32.e_entry;
				break;
			case ELFCLASS64:
				// NOT YET SUPPORTED
				Q_ASSERT(0);
				return header.elf64.e_entry;
				break;
			}
		}
	}
	return 0;
}

//------------------------------------------------------------------------------
// Name: setInitialDebuggerState()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::setInitialDebuggerState() {

	updateMenuState(PAUSED);
		
	m_DebugEventServer = new DebugEventServer(this);
	connect(m_DebugEventServer, SIGNAL(debugEvent(DebugEvent)), this, SLOT(debugEventHandler(DebugEvent)));
	m_DebugEventServer->start();
	
	Debugger::symbolManager = new SymbolManager(Debugger::config()->symbolPath);
	Debugger::memoryRegions = new MemoryRegions(Debugger::debuggerBase->pid());
	Debugger::memoryRegions->sync();
	
	const QVector<MemoryRegions::Region> &l = Debugger::memoryRegions->regions();

	Q_ASSERT(m_DataRegions.size() > 0);

	if(l.size() >= 2) {
		m_DataRegions[0]->region = l[1];
	} else {
		m_DataRegions[0]->region = MemoryRegions::Region();
	}
	
	Debugger::currentState.clear();
	
	m_StepWasRun = false;
}

//------------------------------------------------------------------------------
// Name: setInitialBreakpoint(const QString &s)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::setInitialBreakpoint(const QString &s) {
	BaseTypes::address_t entryPoint = 0;
	
	if(Debugger::config()->initialBreakPoint == Configuration::MainSymbol) {
		const QString mainSymbol = Debugger::basename(s) + "::main";
		const SymbolManager::Symbol *const sym = Debugger::symbolManager->findSymbol(mainSymbol);

		if(sym != 0) {
			entryPoint = sym->address;
		} else {
			entryPoint = getELFEntryPoint(s);
			
			// TODO: make this an independant function
			if(Debugger::config()->findMain) {
			
				ByteShiftArray ba(6);
				for(int i = 0; i < 50; ++i) {
				
					uint8_t byte;
					if(Debugger::debuggerBase->readBytes(entryPoint + i, &byte, sizeof(byte))) {
						ba.shl();
						ba[ba.size() - 1] = byte;
						
						// begining of a call preceeded by a push
						
						// TODO: portability warning!!!
						if(ba[0] == 0x68 && ba[5] == 0xe8) {
							entryPoint = *reinterpret_cast<const uint32_t *>(ba.data() + 1);
							Debugger::log(QString().sprintf("No main symbol found, calculated it to be %p using heuristic", reinterpret_cast<void *>(entryPoint)));
							break;
						}
						
					} else {
						break;
					}
				}
			}
			
		}
	} else if(Debugger::config()->initialBreakPoint == Configuration::EntryPoint) {
		entryPoint = getELFEntryPoint(s);
	}

	if(entryPoint != 0) {
		Debugger::debuggerBase->addBreakpoint(entryPoint);
	}
}

//------------------------------------------------------------------------------
// Name: doOpenFile(const QString &s)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doOpenFile(const QString &s) {
	if(!s.isEmpty()) {
		m_LastOpenDirectory = QFileInfo(s).canonicalFilePath();
	
		on_action_Detach_triggered();

		QStringList args = m_Arguments.split(" ", QString::SkipEmptyParts);
	
		if(Debugger::debuggerBase->open(s, m_WorkingDirectory, args)) {
			setInitialDebuggerState();
			setInitialBreakpoint(s);
			setWindowTitle(QString().sprintf("EDB - %s [%d]", qPrintable(s), Debugger::debuggerBase->pid()));
			
			// update recent file list
			m_RecentFileManager->addFile(s);
			
		} else {
			QMessageBox::information(this, "Could Not Open", "Failed to open and attach to process, please check priviledges and try again.");
		}
		
		Debugger::debuggerBase->getState(Debugger::currentState);
		updateGUI();
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Open_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Open_triggered() {
	doOpenFile(QFileDialog::getOpenFileName(this, "Choose a file", m_LastOpenDirectory, ""));
}

//------------------------------------------------------------------------------
// Name: doAttach(BaseTypes::pid_t pid)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::doAttach(BaseTypes::pid_t pid) {
	if(pid != Debugger::debuggerBase->pid()) {

		on_action_Detach_triggered();

		if(Debugger::debuggerBase->attach(pid)) {
		
			m_WorkingDirectory = Debugger::getProcessCurrentWorkingDirectory();
			QStringList args = Debugger::getProcessArguments();

			setInitialDebuggerState();
			setWindowTitle(QString().sprintf("EDB - %s [%d]", qPrintable(args[0]), pid));
			args.removeFirst();
			
			// make our arguments match what it was run with
			m_Arguments = args.join(" ");
		} else {
			QMessageBox::information(this, "EDB", "Failed to attach to process, please check priviledges and try again.");				
		}

		Debugger::debuggerBase->getState(Debugger::currentState);
		updateGUI();
	} else {
		QMessageBox::information(this, "EDB", "You are already debugging that process!");
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Attach_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Attach_triggered() {
	static DialogAttach *dlg = 0;
	if(dlg == 0) {
		dlg = new DialogAttach(this);
	} 
	
	const int ret = dlg->exec();
	
	if(ret == QDialog::Accepted) {
		QList<QTableWidgetItem *> sel = dlg->ui.tableWidget->selectedItems();
		if(sel.size() != 0) {
			const BaseTypes::pid_t pid = sel.begin()[0]->text().toUInt();
			doAttach(pid);
		}
	}
}

//------------------------------------------------------------------------------
// Name: on_action_Memory_Regions_triggered()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerMain::on_action_Memory_Regions_triggered() {
	static DialogMemoryRegions *dlg = 0;
	if(dlg == 0) {
		dlg = new DialogMemoryRegions(this);
	}
	
	const int ret = dlg->exec();
	
	if(ret == QDialog::Accepted) {
		QList<QTableWidgetItem *> sel = dlg->ui.tableWidget->selectedItems();
		if(sel.size() != 0) {
			const BaseTypes::address_t region_start	= sel.begin()[0]->text().toUInt(0, 16);
			Debugger::dumpData(region_start);
		}
	}
}

//------------------------------------------------------------------------------
// Name: createNewDataTab()
// Desc: duplicates the current tab creating a new one
//------------------------------------------------------------------------------
void DebuggerMain::createNewDataTab() {
	
	const int currentIndex = ui.tabWidget->currentIndex();
	
	DataViewInfo *const newDataView = new DataViewInfo;
	
	if(currentIndex != -1) {
		// duplicate the current region
		newDataView->region = m_DataRegions[currentIndex]->region;
	}
	
	newDataView->stream	= new ByteStreamRegion(newDataView->region);
	newDataView->view	= new QHexView;
		
	// setup the context menu
	newDataView->view->setContextMenuPolicy(Qt::CustomContextMenu);
	connect(newDataView->view, SIGNAL(customContextMenuRequested(const QPoint &)), this, SLOT(dumpContextMenu(const QPoint &)));
	
	newDataView->view->setAddressOffset(newDataView->region.start);
	newDataView->view->setData(newDataView->stream);
	
	// set the default font
	QFont dumpFont;
	dumpFont.fromString(Debugger::config()->MemoryDumpViewFont);
	newDataView->view->setFont(dumpFont);
	
	
	m_DataRegions.push_back(newDataView);

	// create the tab!
	ui.tabWidget->addTab(newDataView->view, QString().sprintf("%08x-%08x", newDataView->region.start, newDataView->region.end));
	ui.tabWidget->setCurrentIndex(ui.tabWidget->count() - 1);
	
	
	m_DelTab->setEnabled(ui.tabWidget->count() > 1);
}

//------------------------------------------------------------------------------
// Name: removeDataTab() 
// Desc: handles removing of a memory view tab
//------------------------------------------------------------------------------
void DebuggerMain::removeDataTab() {
	const int currentIndex = ui.tabWidget->currentIndex();
	
	DataViewInfo *info = m_DataRegions[currentIndex];
	m_DataRegions.remove(currentIndex);
	
	ui.tabWidget->removeTab(currentIndex);
	
	m_DelTab->setEnabled(ui.tabWidget->count() > 1);
	
	if(info != 0) {
		safeDelete(info->stream);
		safeDelete(info->view);
	}
	
	safeDelete(info);
}


//------------------------------------------------------------------------------
// Name: valueFromName(const QString &s, unsigned int mode)
// Desc: gets the value of a register based on name, takes into account a "mode"
//       mask, which lets us select catagories of registers to consider valid
// Note: portability issue, implementation uses x86 specific names
//------------------------------------------------------------------------------
BaseTypes::reg_t *DebuggerMain::valueFromName(const QString &s, unsigned int mode) {

	// TODO: move this elsewhere, it doesn't really need to be here!
	if(mode & State::REG_COND && s.startsWith("EFLAGS:"))	return &Debugger::currentState.eflags;
	if(mode & State::REG_GPR && s.startsWith("EAX:")) 		return &Debugger::currentState.eax;
	if(mode & State::REG_GPR && s.startsWith("EBX:")) 		return &Debugger::currentState.ebx;
	if(mode & State::REG_GPR && s.startsWith("ECX:")) 		return &Debugger::currentState.ecx;
	if(mode & State::REG_GPR && s.startsWith("EDX:")) 		return &Debugger::currentState.edx;
	if(mode & State::REG_GPR && s.startsWith("EBP:")) 		return &Debugger::currentState.ebp;
	if(mode & State::REG_GPR && s.startsWith("ESP:")) 		return &Debugger::currentState.esp;
	if(mode & State::REG_GPR && s.startsWith("EDI:")) 		return &Debugger::currentState.edi;;
	if(mode & State::REG_GPR && s.startsWith("ESI:")) 		return &Debugger::currentState.esi;
	if(mode & State::REG_IP && s.startsWith("EIP:")) 		return &Debugger::currentState.eip;
	if(mode & State::REG_SEG && s.startsWith("CS:")) 		return &Debugger::currentState._cs;
	if(mode & State::REG_SEG && s.startsWith("DS:")) 		return &Debugger::currentState._ds;
	if(mode & State::REG_SEG && s.startsWith("ES:")) 		return &Debugger::currentState._es;
	if(mode & State::REG_SEG && s.startsWith("FS:")) 		return &Debugger::currentState._fs;
	if(mode & State::REG_SEG && s.startsWith("GS:")) 		return &Debugger::currentState._gs;
	if(mode & State::REG_SEG && s.startsWith("SS:")) 		return &Debugger::currentState._ss;
	
	return 0;
}
