/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DisassemblyLib.h"
#include "State.h"
#include <cstring>

// this is a kludge to gaurantee that init and cleanup are called
static struct GlobalDisassemblyInit {
	GlobalDisassemblyInit()	{ 
		x86_init(opt_none, 0, 0);
	}
	
	~GlobalDisassemblyInit()	{ 
		x86_cleanup();
	}
	
} global_disassembly;


//------------------------------------------------------------------------------
// Name: Instruction(const uint8_t *buffer, std::size_t size, BaseTypes::address_t rva)
//------------------------------------------------------------------------------
Instruction::Instruction(const uint8_t *buffer, std::size_t size, BaseTypes::address_t rva) {

	x86_disasm(
			const_cast<uint8_t *>(buffer),
			size, 
			rva, 
			0, 
			&m_Internal_instruction
			);
}

//------------------------------------------------------------------------------
// Name: ~Instruction()
//------------------------------------------------------------------------------
Instruction::~Instruction() {
	if(valid()) {
		x86_oplist_free(&m_Internal_instruction);
	}
}

//------------------------------------------------------------------------------
// Name: format(char *buf, std::size_t len, Configuration::Syntax syntax) const
//------------------------------------------------------------------------------
void Instruction::format(char *buf, std::size_t len, Configuration::Syntax syntax) const {
	if (valid()) {
		if(syntax == Configuration::Intel) {
			x86_format_insn(const_cast<x86_insn_t *>(&m_Internal_instruction), buf, len, intel_syntax);
		} else {
			x86_format_insn(const_cast<x86_insn_t *>(&m_Internal_instruction), buf, len, att_syntax);
		}
	} else {
		const uint8_t byte = m_Internal_instruction.bytes[0] & 0xff;
		if(syntax == Configuration::Intel) {
			snprintf(buf, len, "db %02X", byte);
		} else {
			snprintf(buf, len, ".byte %#02x", byte);
		}
	}
}

//------------------------------------------------------------------------------
// Name: format(char *buf, std::size_t len, Configuration::Syntax syntax) const
//------------------------------------------------------------------------------
void Operand::format(char *buf, std::size_t len, Configuration::Syntax syntax) const {
	
	if(syntax == Configuration::Intel) {
		x86_format_operand(impl(), buf, len, intel_syntax);
	} else {
		x86_format_operand(impl(), buf, len, att_syntax);
	}
}

//------------------------------------------------------------------------------
// Name: effectiveAddress(const State &state) const
//------------------------------------------------------------------------------
BaseTypes::address_t Operand::effectiveAddress(const State &state) const {
	
	const x86_op_t *const op = impl();
	const x86_insn_t *const owner = reinterpret_cast<x86_insn_t *>(op->insn);
	
	Q_ASSERT(owner != 0);
	
	uint32_t ret = 0;
	switch(type()) {
	case REGISTER:
		ret = state.value(op->data.reg.name);
		break;		
	case RELATIVE_NEAR:
		ret = owner->addr + owner->size + op->data.relative_near;
		break;	
	case RELATIVE_FAR:
		ret = owner->addr + owner->size + op->data.relative_far;
		break;	
	case ABSOLUTE:
		// TODO: find a way to take into account the segment base/limit
		ret = op->data.absolute.offset.off32;
		break;
	case OFFSET:
		// TODO: find a way to take into account the segment base/limit
		ret = op->data.offset;
		break;
	case EXPRESSION:
		do {
			const BaseTypes::reg_t base		= state.value(op->data.expression.base.name);
			const BaseTypes::reg_t index	= state.value(op->data.expression.index.name);
			ret								= base + index * op->data.expression.scale + op->data.expression.disp;
		} while(0);
		break;
	
	case IMMEDIATE:
		break;
	default:
		break;
	}
	
	return ret;
}
