/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef DISASSEMBLYLIB_06122006_H_
#define DISASSEMBLYLIB_06122006_H_

#include <libdis.h>
#include <stdint.h>
#include <cstddef>
#include "types.h"
#include "Configuration.h"

class State;

class OperandBase {
public:
	virtual ~OperandBase() {}
	
protected:
	friend class Instruction;
	OperandBase(void *data) : m_Data(data) {}
	
public:
	enum Type {
        UNUSED			= 0,	// empty/unused operand: should never occur
        REGISTER		= 1,	// CPU register
        IMMEDIATE		= 2,	// Immediate Value
        RELATIVE_NEAR	= 3,  	// Relative offset from IP
        RELATIVE_FAR	= 4,   	// Relative offset from IP
        ABSOLUTE		= 5,	// Absolute address (ptr16:32)
        EXPRESSION		= 6,	// Address expression (scale/index/base/disp)
        OFFSET			= 7,	// Offset from start of segment (m32)
        UNKNOWN
	};

public:
	virtual BaseTypes::address_t effectiveAddress(const State &state) const = 0;
	virtual Type type() const = 0;
	virtual bool valid() const = 0;
	virtual void format(char *buf, std::size_t len, Configuration::Syntax syntax) const = 0;

protected:
	void *const m_Data;
};


class Operand : public OperandBase {
protected:
	friend class Instruction;
	Operand(void *data) : OperandBase(data) {}
	
public:
	void format(char *buf, std::size_t len, Configuration::Syntax syntax) const;
	virtual BaseTypes::address_t effectiveAddress(const State &state) const;
	virtual Type type() const	{ return static_cast<Type>(impl()->type); }
	virtual bool valid() const	{ return m_Data != 0; }
	
private:
	x86_op_t *impl() const { return reinterpret_cast<x86_op_t *>(m_Data); }

};

class Instruction {
public:
	enum Prefix {
		PREFIX_NONE				= 0x0000,
		PREFIX_LOCK				= 0x0001,
		PREFIX_REPNE			= 0x0002,
		PREFIX_REP				= 0x0004,
		PREFIX_CS				= 0x0008,
		PREFIX_SS				= 0x0010,
		PREFIX_DS				= 0x0020,
		PREFIX_ES				= 0x0040,
		PREFIX_FS				= 0x0080,
		PREFIX_GS				= 0x0100,
		PREFIX_BRANCH_NOT_TAKEN	= 0x0200,
		PREFIX_BRANCH_TAKEN		= 0x0400,
		PREFIX_OPERAND			= 0x0800,
		PREFIX_ADDRESS			= 0x1000
	};
public:
	Instruction(const uint8_t *buffer, std::size_t size, BaseTypes::address_t rva);
	~Instruction();

private:
	// not copyable
	Instruction(const Instruction&);
	Instruction &operator=(const Instruction&);
	
public:
	const Operand operand_1st() const 	{ return Operand(x86_operand_1st(const_cast<x86_insn_t *>(&m_Internal_instruction))); }
	const Operand operand_2nd() const	{ return Operand(x86_operand_2nd(const_cast<x86_insn_t *>(&m_Internal_instruction))); }
	const Operand operand_3rd() const	{ return Operand(x86_operand_3rd(const_cast<x86_insn_t *>(&m_Internal_instruction))); }
	
	uint16_t prefix() const				{ return m_Internal_instruction.prefix; }
	x86_insn_type type() const			{ return m_Internal_instruction.type; }
	x86_flag_status flagsTested() const	{ return m_Internal_instruction.flags_tested; }
	int size() const					{ return m_Internal_instruction.size; }
	const uint8_t *data() const			{ return m_Internal_instruction.bytes; }
	bool valid() const					{ return type() != insn_invalid; }
	
	void format(char *buf, std::size_t len, Configuration::Syntax syntax) const;

private:
	x86_insn_t m_Internal_instruction;
};

#endif
