/*
Copyright (C) 2006 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "i386ArchProcessor.h"
#include "QCatagoryList.h"
#include "Debugger.h"
#include "DebuggerCoreInterface.h"
#include "DisassemblyLib.h"
#include "State.h"
#include "Util.h"

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
i386ArchProcessor::i386ArchProcessor() {

	const struct Function {
		const char *name;
		FunctionInfo f;
	} funcs[] = {
		// initialize our system with some libc definitions
		{ "memchr", FunctionInfo('p', 'c', 'u') },
		{ "memcmp", FunctionInfo('p', 'p', 'u') },
		{ "memcpy", FunctionInfo('p', 'p', 'u') },
		{ "memmove", FunctionInfo('p', 'p', 'u') },
		{ "memset", FunctionInfo('p', 'c', 'u') },
		{ "strcat", FunctionInfo('s', 's') },
		{ "strchr", FunctionInfo('s', 'c') },
		{ "strcmp", FunctionInfo('s', 's') },
		{ "strcoll", FunctionInfo('s', 's') },
		{ "strcpy", FunctionInfo('p', 's') },
		{ "strcspn", FunctionInfo('s', 's') },
		{ "strerror", FunctionInfo('i') },
		{ "strlen", FunctionInfo('s') },
		{ "strncat", FunctionInfo('s', 's', 'u') },
		{ "strncmp", FunctionInfo('s', 's', 'u') },
		{ "strncpy", FunctionInfo('p', 's', 'u') },
		{ "strpbrk", FunctionInfo('s', 's') },
		{ "strrchr", FunctionInfo('s', 'c') },
		{ "strspn", FunctionInfo('s', 's') },
		{ "strstr", FunctionInfo('s', 's') },
		{ "strtok", FunctionInfo('s', 's') },
		{ "strxfrm", FunctionInfo('p', 's', 'u') },
		{ "strdup", FunctionInfo('s') },
		{ "strndup", FunctionInfo('s', 'u') },

		{ "fopen", FunctionInfo('s', 'p') },
		{ "fclose", FunctionInfo('p') },
		{ "fread", FunctionInfo('p', 'u', 'u', 'p') },
		{ "fwrite", FunctionInfo('p', 'u', 'u', 'p') },

		{ "exit", FunctionInfo('i') },
		{ "printf", FunctionInfo('f') },
		{ "fputc", FunctionInfo('c', 'p') },
		{ "fputs", FunctionInfo('s', 'p') },
		{ "putchar", FunctionInfo('c') },
		{ "puts", FunctionInfo('p') },
		{ "remove", FunctionInfo('s') },
		{ "setlocale", FunctionInfo('i', 's') },

		// other functions
		{ "__errno_location", FunctionInfo() },
		{ "__libc_start_main", FunctionInfo('p', 'i', 'p', 'p', 'p', 'p', 'p') },
		{ "getpagesize", FunctionInfo() },
		{ "ioctl", FunctionInfo('i', 'i', 'p') },
		{ "isatty", FunctionInfo('i') },
		{ "poll", FunctionInfo('p', 'u', 'i') },
		{ "getenv", FunctionInfo('s') },
		{ "setenv", FunctionInfo('s', 's', 'i') },
		{ "bindtextdomain", FunctionInfo('s', 's') },
		{ "textdomain", FunctionInfo('s') },
		{ "getopt", FunctionInfo('i', 'p', 's') },
		{ "getopt_long", FunctionInfo('i', 'p', 's', 'p', 'p') },
		{ 0, FunctionInfo() }
	};
	
	const Function *p = funcs;
	while(p->name != 0) {
		m_FunctionInfo.insert(p->name, p->f);
		++p;
	}



}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::setupRegisterView(QCatagoryList *catagoryList) {

	Q_ASSERT(catagoryList != 0);

	// setup the register view
	QTreeWidgetItem *const gpr = catagoryList->addCatagory("General Purpose");
	
	Q_ASSERT(gpr != 0);
	
	for(int i = 0; i < 10; ++i) {
		m_RegisterViewItems.push_back(catagoryList->addItem(gpr, ""));
	}
	
	QTreeWidgetItem *const segs = catagoryList->addCatagory("Segments");
	
	Q_ASSERT(segs != 0);
	
	for(int i = 0; i < 6; ++i) {
		m_RegisterViewItems.push_back(catagoryList->addItem(segs, ""));
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QTreeWidgetItem *i386ArchProcessor::getRegisterItem(unsigned int index) {
	QTreeWidgetItem *ret = 0;
	if(index < static_cast<unsigned int>(m_RegisterViewItems.size())) {
		ret = m_RegisterViewItems[index];
	}
	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::doUpdateRegister(QTreeWidgetItem *item, const QString &name, BaseTypes::reg_t value) {
	
	Q_ASSERT(item != 0);

	QString regString;
	// TODO: uses elided text for ascii string?
	if(Debugger::memoryRegions != 0 && Debugger::getASCIIStringAtAddress(value, regString, Debugger::config()->minAsciiStringLength, 128)) {
		item->setText(0, QString().sprintf("%s: %08x ASCII \"%s\"", qPrintable(name), value, qPrintable(regString)));
	} else {
		item->setText(0, QString().sprintf("%s: %08x", qPrintable(name), value));
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::updateRegisterView(const QString &defaultRegionName) {
	if(Debugger::memoryRegions != 0) {
		doUpdateRegister(getRegisterItem(0), "EAX", Debugger::currentState.eax);
		doUpdateRegister(getRegisterItem(1), "EBX", Debugger::currentState.ebx);
		doUpdateRegister(getRegisterItem(2), "ECX", Debugger::currentState.ecx);
		doUpdateRegister(getRegisterItem(3), "EDX", Debugger::currentState.edx);
		doUpdateRegister(getRegisterItem(4), "EBP", Debugger::currentState.ebp);
		doUpdateRegister(getRegisterItem(5), "ESP", Debugger::currentState.esp);
		doUpdateRegister(getRegisterItem(6), "ESI", Debugger::currentState.esi);
		doUpdateRegister(getRegisterItem(7), "EDI", Debugger::currentState.edi);
		
		const QString symname = Debugger::findFunctionSymbol(Debugger::currentState.eip, defaultRegionName);
		getRegisterItem(8)->setText(0, QString().sprintf("EIP: %08x <%s>", Debugger::currentState.eip, qPrintable(symname)));
		getRegisterItem(9)->setText(0, QString().sprintf("EFLAGS: %08x", Debugger::currentState.eflags));

		getRegisterItem(10)->setText(0, QString().sprintf("CS: %04x", Debugger::currentState.cs & 0xffff));
		getRegisterItem(11)->setText(0, QString().sprintf("DS: %04x", Debugger::currentState.ds & 0xffff));
		getRegisterItem(12)->setText(0, QString().sprintf("ES: %04x", Debugger::currentState.es & 0xffff));
		getRegisterItem(13)->setText(0, QString().sprintf("FS: %04x", Debugger::currentState.fs & 0xffff));
		getRegisterItem(14)->setText(0, QString().sprintf("GS: %04x", Debugger::currentState.gs & 0xffff));
		getRegisterItem(15)->setText(0, QString().sprintf("SS: %04x", Debugger::currentState.ss & 0xffff));
		
		
		getRegisterItem(0)->setForeground(0, QBrush((Debugger::currentState.eax != m_LastState.eax) ? Qt::red : Qt::black));
		getRegisterItem(1)->setForeground(0, QBrush((Debugger::currentState.ebx != m_LastState.ebx) ? Qt::red : Qt::black));
		getRegisterItem(2)->setForeground(0, QBrush((Debugger::currentState.ecx != m_LastState.ecx) ? Qt::red : Qt::black));
		getRegisterItem(3)->setForeground(0, QBrush((Debugger::currentState.edx != m_LastState.edx) ? Qt::red : Qt::black));
		getRegisterItem(4)->setForeground(0, QBrush((Debugger::currentState.ebp != m_LastState.ebp) ? Qt::red : Qt::black));
		getRegisterItem(5)->setForeground(0, QBrush((Debugger::currentState.esp != m_LastState.esp) ? Qt::red : Qt::black));
		getRegisterItem(6)->setForeground(0, QBrush((Debugger::currentState.esi != m_LastState.esi) ? Qt::red : Qt::black));
		getRegisterItem(7)->setForeground(0, QBrush((Debugger::currentState.edi != m_LastState.edi) ? Qt::red : Qt::black));
		getRegisterItem(8)->setForeground(0, QBrush((Debugger::currentState.eip != m_LastState.eip) ? Qt::red : Qt::black));
		getRegisterItem(9)->setForeground(0, QBrush((Debugger::currentState.eflags != m_LastState.eflags) ? Qt::red : Qt::black));

		m_LastState = Debugger::currentState;
	} else {
		getRegisterItem(0)->setText(0, "EAX: 00000000");
		getRegisterItem(1)->setText(0, "EBX: 00000000");
		getRegisterItem(2)->setText(0, "ECX: 00000000");
		getRegisterItem(3)->setText(0, "EDX: 00000000");
		getRegisterItem(4)->setText(0, "EBP: 00000000");
		getRegisterItem(5)->setText(0, "ESP: 00000000");
		getRegisterItem(6)->setText(0, "ESI: 00000000");
		getRegisterItem(7)->setText(0, "EDI: 00000000");
		getRegisterItem(8)->setText(0, "EIP: 00000000");
		getRegisterItem(9)->setText(0, "EFLAGS: 00000000");
		getRegisterItem(10)->setText(0, "CS: 0000");
		getRegisterItem(11)->setText(0, "DS: 0000");
		getRegisterItem(12)->setText(0, "ES: 0000");
		getRegisterItem(13)->setText(0, "FS: 0000");
		getRegisterItem(14)->setText(0, "GS: 0000");
		getRegisterItem(15)->setText(0, "SS: 0000");
		
		for(int i = 0; i < 15; ++i) {
			getRegisterItem(7)->setForeground(0, QBrush(Qt::black));
		}
		
		m_LastState = State();
	}
	
	
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QStringList i386ArchProcessor::doFunctionParamterResolution(const QString &symname, int offset) {

	QStringList ret;

	// we will always be removing the last 2 chars "+0" from the string as well
	// as chopping the region prefix we like to prepend to symbols
	QString funcName;
	const int indexColon = symname.indexOf("::");

	if(indexColon != -1) {
		funcName = symname.left(symname.length() - 2).mid(indexColon + 2);
	}

	// safe not to check for -1, it means "rest of string" for the mid function
	funcName = funcName.mid(0, funcName.indexOf("@"));
	
	QMap<QString, FunctionInfo>::const_iterator it = m_FunctionInfo.find(funcName);
	if(it != m_FunctionInfo.end()) {
	
		QString functionCall = funcName + "(";
		
		const QVector<QChar> params(it->params());
		int i = 0;
		foreach(QChar ch, params) {
		
			BaseTypes::reg_t arg;
			Debugger::debuggerBase->readBytes(Debugger::currentState.esp + i * sizeof(BaseTypes::reg_t) + offset, &arg, sizeof(arg));
			
			switch(ch.toAscii()) {
			case 's':
			case 'f':
				do {
					QString stringParam;
					if(Debugger::getASCIIStringAtAddress(arg, stringParam, Debugger::config()->minAsciiStringLength, INT_MAX)) {
						functionCall += QString().sprintf("<0x%08x> \"%s\"", arg, qPrintable(stringParam));
					} else {
						char character;
						Debugger::debuggerBase->readBytes(arg, &character, sizeof(character));
						if(character == '\0') {
							functionCall += QString().sprintf("<0x%08x> \"\"", arg);
						} else {						
							functionCall += QString().sprintf("0x%08x", arg);
						}
					}
				} while(0);
				break;
					
			case 'p':
				functionCall += QString().sprintf("0x%08x", arg);
				break;
			case 'c':
				if(safe_ctype<std::isprint>(arg) || arg == ' ') {
					functionCall += QString().sprintf("'%c'", arg);
				} else {
					functionCall += QString().sprintf("'\\x%02x'", arg);
				}
				break;
			case 'u':
				functionCall += QString().sprintf("%u", arg);
				break;
			case 'i':
				functionCall += QString().sprintf("%d", arg);
				break;
			}
			
			if(i + 1 < params.size()) {
				functionCall += ", ";
			}
			++i;
		}
		
		functionCall += ")";
		ret.push_back(functionCall);
	}

	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::analyzeJump(const Operand *operands, const Instruction &insn, QStringList &ret) {

	Q_UNUSED(operands);

	bool taken = true;
	const bool cf = (Debugger::currentState.eflags & 0x0001) != 0;
	const bool pf = (Debugger::currentState.eflags & 0x0004) != 0;
	const bool zf = (Debugger::currentState.eflags & 0x0040) != 0;
	const bool sf = (Debugger::currentState.eflags & 0x0080) != 0;
	const bool df = (Debugger::currentState.eflags & 0x0400) != 0;
	const bool of = (Debugger::currentState.eflags & 0x0800) != 0;

	if((insn.flagsTested() & insn_carry_set)				&& !cf) taken = false;
	if((insn.flagsTested() & insn_zero_set)					&& !zf) taken = false;
	if((insn.flagsTested() & insn_oflow_set)				&& !of) taken = false;
	if((insn.flagsTested() & insn_dir_set)					&& !df) taken = false;
	if((insn.flagsTested() & insn_sign_set)					&& !sf) taken = false;
	if((insn.flagsTested() & insn_parity_set)				&& !pf) taken = false;
	if((insn.flagsTested() & insn_carry_clear)				&& cf) taken = false;
	if((insn.flagsTested() & insn_zero_clear)				&& zf) taken = false;
	if((insn.flagsTested() & insn_oflow_clear)				&& of) taken = false;
	if((insn.flagsTested() & insn_dir_clear)				&& df) taken = false;
	if((insn.flagsTested() & insn_sign_clear)				&& sf) taken = false;
	if((insn.flagsTested() & insn_parity_clear)				&& pf) taken = false;

	if((insn.flagsTested() & insn_zero_set_or_sign_ne_oflow)&& (!zf && !(sf != of)))	taken = false;
	if((insn.flagsTested() & insn_carry_or_zero_set)		&& !(cf && zf))				taken = false;
	if((insn.flagsTested() & insn_sign_eq_oflow)			&& !(sf == of))				taken = false;
	if((insn.flagsTested() & insn_sign_ne_oflow)			&& (sf == of))				taken = false;

	if(taken) {
		ret.push_back("jump taken");
	} else {
		ret.push_back("jump NOT taken");
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::analyzeReturn(const Operand *operands, const Instruction &insn, QStringList &ret) {

	Q_UNUSED(operands);
	Q_UNUSED(insn);

	BaseTypes::address_t returnAddress;
	Debugger::debuggerBase->readBytes(Debugger::currentState.esp, &returnAddress, sizeof(returnAddress));

	const QString symname = Debugger::findFunctionSymbol(returnAddress, "");
	if(!symname.isEmpty()) {
		ret.push_back(QString().sprintf("Return to %08x <%s>", returnAddress, qPrintable(symname)));
	} else {
		ret.push_back(QString().sprintf("Return to %08x", returnAddress));
	}
}


//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::analyzeCall(const Operand *operands, const Instruction &insn, QStringList &ret) {
	
	Q_ASSERT(operands != 0);
	
	if(operands[0].valid()) {

		const BaseTypes::address_t effective_address = operands[0].effectiveAddress(Debugger::currentState);
		char tmp[100];

		operands[0].format(tmp, sizeof(tmp), Debugger::config()->syntax);

		switch(operands[0].type()) {
		case op_relative_far:
		case op_relative_near:
		case op_register:
			do {
				int offset;
				const QString symname = Debugger::findFunctionSymbol(effective_address, "", &offset);
				if(!symname.isEmpty()) {
					ret.push_back(QString().sprintf("%s = %08x <%s>", tmp, effective_address, qPrintable(symname)));

					if(offset == 0) {
						if(insn.type() == insn_call) {
							ret += doFunctionParamterResolution(symname, 0);
						} else {
							ret += doFunctionParamterResolution(symname, 4);
						}
					}

				} else {
					ret.push_back(QString().sprintf("%s = %08x", tmp, effective_address));
				}
			} while(0);
			break;

		case op_expression:
		default:
			do {
				BaseTypes::address_t target;
				const bool ok = Debugger::debuggerBase->readBytes(effective_address, &target, sizeof(target));

				if(ok) {
					int offset;
					const QString symname = Debugger::findFunctionSymbol(target, "", &offset);
					if(!symname.isEmpty()) {
						ret.push_back(QString().sprintf("%s = [%08x] = %08x <%s>", tmp, effective_address, target, qPrintable(symname)));

						if(offset == 0) {
							if(insn.type() == insn_call) {
								ret += doFunctionParamterResolution(symname, 0);
							} else {
								ret += doFunctionParamterResolution(symname, 4);
							}
						}

					} else {
						ret.push_back(QString().sprintf("%s = [%08x] = %08x", tmp, effective_address, target));
					}
				} else {
					// could not read from the address
					ret.push_back(QString().sprintf("%s = [%08x] = ?", tmp, effective_address));
				}
			} while(0);
			break;
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void i386ArchProcessor::analyzeOperands(const Operand *operands, const Instruction &insn, QStringList &ret) {

	Q_UNUSED(insn);
	Q_ASSERT(operands != 0);

	for(int j = 0; j < 3; ++j) {

		if(operands[j].valid()) {

			char tmp[100];
			operands[j].format(tmp, sizeof(tmp), Debugger::config()->syntax);

			switch(operands[j].type()) {
			case op_relative_far:
			case op_relative_near:
			case op_register:
				do {
					const BaseTypes::address_t effective_address = operands[j].effectiveAddress(Debugger::currentState);
					ret.push_back(QString().sprintf("%s = %08x", tmp, effective_address));
				} while(0);
				break;
			case op_expression:
				do {
					const BaseTypes::address_t effective_address = operands[j].effectiveAddress(Debugger::currentState);
					BaseTypes::address_t target;

					const bool ok = Debugger::debuggerBase->readBytes(effective_address, &target, sizeof(target));

					if(ok) {
						ret.push_back(QString().sprintf("%s = [%08x] = %08x", tmp, effective_address, target));
					} else {
						ret.push_back(QString().sprintf("%s = [%08x] = ?", tmp, effective_address));
					}
				} while(0);
				break;
			default:
				break;
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QStringList i386ArchProcessor::updateInstructionInfo(BaseTypes::address_t address) {

	QStringList ret;
	
	if(Debugger::debuggerBase != 0) {
		uint8_t buffer[MAX_INSN_SIZE];
		
		const bool ok = Debugger::debuggerBase->readBytes(address, buffer, sizeof(buffer));
		if(ok) {
			Instruction insn(buffer, sizeof(buffer), address);
			if(insn.valid()) {

				// fetch the 3 possible operands
				const Operand operands[3] = {
					insn.operand_1st(),
					insn.operand_2nd(),
					insn.operand_3rd()
				};

				// figure out the instruction type and display some information about it
				switch(insn.type()) {
				case insn_return:
					analyzeReturn(operands, insn, ret);
					break;
				case insn_jcc:
					analyzeJump(operands, insn, ret);
					// FALL THROUGH!
				case insn_jmp:
				case insn_call:
					analyzeCall(operands, insn, ret);
					break;
				default:
					analyzeOperands(operands, insn, ret);
					break;
				}
			}
		}
	}
	
	// eliminate duplicates
	ret = QStringList::fromSet(ret.toSet());
	
	return ret;
}
